/*
 * Created on 2005-03-08
 *
 */
package vr.system.classifiersManagement;

import java.io.File;
import java.util.Collection;

import rseslib.processing.classification.AbstractMultiClassifier;
import rseslib.structure.attribute.ArrayHeader;
import rseslib.structure.attribute.Header;
import rseslib.structure.attribute.formats.DataFormatRecognizer;
import rseslib.structure.data.ArrayListDoubleDataTable;
import rseslib.structure.data.DoubleDataTable;
import rseslib.system.ConfigurationWithStatistics;
import rseslib.system.Report;
import rseslib.system.progress.EmptyProgress;
import vr.processing.classification.GeneralMultiClassifier;
import vr.system.DataManager;

/**
 * Klasa odpowiedzialna za nieinterfejsowa czesc rejestracji klasyfikatorow.
 * Przetestowanie ich zgodnosci z systemem.
 * Wykrycie wlasnosci klasyfikatorow (czy multikasyfikator, na jakich danych dziala)
 * lub sprawdzenie zgodnosci z deklaracja
 *
 * @author Marcin Staszczyk
 */
public class ClassifierRegistrator {

	/**
	 * Flagi klasyfikatorow
	 */
	/** Klasyfikator jest multiklasyfikatorem */
	public static final int FLAG_MULTICLASSIFIER = 1;
	/** Klasyfikator powinien sobie radzic z danymi o argumentach numerycznych i decyzji numerycznej */
	public static final int FLAG_NUMERIC_TO_NUMERIC = 2;
	/** Klasyfikator powinien sobie radzic z danymi o argumentach numerycznych i decyzji symbolicznej */
	public static final int FLAG_NUMERIC_TO_NOMINAL = 4;
	/** Klasyfikator powinien sobie radzic z danymi o argumentach symbolicznych i decyzji numerycznej */
	public static final int FLAG_NOMINAL_TO_NUMERIC = 8;
	/** Klasyfikator powinien sobie radzic z danymi o argumentach symbolicznych i decyzji symbolicznej */
	public static final int FLAG_NOMINAL_TO_NOMINAL = 16;
	/** Rejestracja klasyfikatora ma sama rozpoznac flagi */
	public static final int FLAG_TRY_TO_DETERMINE_FLAGS = 32;

	/**
	 * Etapy rejestracji (stany w ktorych znajduje sie Rejestrator)
	 */
	private static final int CLASS_CREATED = 0;
	private static final int CLASSIFIER_FILES_ADDED = 1;
	private static final int PROPERTY_FILES_ADDED = 2;
	private static final int ALL_DONE = 3;
	private static final int ERROR = -1;

	/** stan klasyfikatora */
	private int state = CLASS_CREATED;

	/** Scierzka java-owa do podanego pliku .class */
	private String classifierJavaPathWithoutName;
	/** Nazwa nowego klasyfikatora */
	private String classifierName;
	/** Flagi nowego klasyfikatora */
	private int newFlags;
	/** Obiekt Class utoworzony na podstawie podanego pliku *.class */
	private Class classClassifier;

	/**
	 * Metoda odpowiedzialna za wykonanie operacji sprawdzajacych
	 * podstawowye inforamcje o pliku i zapisanie ich w pamieci dla dalszych metod
	 *
	 * @param classifier 		Java-owa scierzka do pliku z klasyfikatorem, ktora podana metodzie Class.forName nie zwroci zadnych bladow
	 * @param flags				Flagi z jakimi wczytano pliki (parz pierwsza grupa stalych intowych)
	 * @return					zwraca, czy klasyfikator jest juz na liscie zarejestrowanych
	 * @throws RuntimeException jesli wystapily jakies bledy
	 */
	public boolean classifier(String classifier, int flags) throws RuntimeException
	{
		if (state != CLASS_CREATED)
		{
			throw new RuntimeException("Uruchomiono dodawanie klasyfikatora w nieodpowiednim momencie");
		}

		if ((flags & ~FLAG_MULTICLASSIFIER) == 0)
		{
			throw new RuntimeException("Klasyfikator jest bezuzyteczny");
		}
		int sub = classifier.lastIndexOf('.');
		if ((sub < 0) || (sub > classifier.length()))
		{
			classifierJavaPathWithoutName = "";
			classifierName = classifier;
		}
		else
		{
			classifierName = classifier.substring(sub+1);
			classifierJavaPathWithoutName = classifier.substring(0, sub+1);
		}

		String newClassifierAcces = classifierJavaPathWithoutName + classifierName;
		newFlags = flags;

		boolean result = testClassifier(newClassifierAcces, classifierName, newFlags);

		state = CLASSIFIER_FILES_ADDED;

		return result;
	}

	/**
	 * Tworzy z obiektu typu file scierzke wzgledna od katalogu projektu
	 * do pliku z klasa w postaci java-owej.
	 * Odcina pierwsza czesc scierzki bedaca scierzka do aktualnego katalogu.
	 *
	 * @param classifier	plik z klasa
	 * @return				scierzka wzgledna od katalogu projektu
	 * do pliku z klasa w postaci java-owej
	 */
	private String makeJavaPath(File classifier)
	{
		return makeJavaDir(classifier) + makeJavaName(classifier);
	}

	/**
	 * Tworzy z obiektu typu file scierzke wzgledna od katalogu projektu
	 * do pliku z klasa w postaci java-owej (czesc odpowiadajaca za przygotowanie
	 * czesci katalogowej).
	 * Odcina pierwsza czesc scierzki bedaca scierzka do aktualnego katalogu.
	 *
	 * @param classifier	plik z klasa
	 * @return				scierzka wzgledna od katalogu projektu
	 * do pliku z klasa w postaci java-owej
	 */
	private String makeJavaDir(File classifier)
	{
		String workDirName = new File("").getAbsolutePath();

		if (! classifier.getAbsolutePath().startsWith(workDirName))
			throw new RuntimeException("BLAD. Plik docelowy nie w folderze roboczym VR");

		int workDirLength = workDirName.length();

		String pom = classifier.getAbsolutePath().substring(workDirLength+1).replace(File.separatorChar,'.');

		return pom.substring(0,pom.length() - classifier.getName().length()); /* usuwamy "nazwa.class" */
	}

	/**
	 * Tworzy z obiektu typu file scierzke wzgledna od katalogu projektu
	 * do pliku z klasa w postaci java-owej (czesc odpowiadajaca za przygotowanie
	 * czesci z nazwa klasy)
	 *
	 * @param classifier	plik z klasa
	 * @return				scierzka wzgledna od katalogu projektu
	 * do pliku z klasa w postaci java-owej
	 */
	private String makeJavaName(File classifier)
	{
		String pom = classifier.getName();

		return pom.substring(0,pom.length()-6); /* usuwamy ".class" */
	}

	/**
	 * Metoda odpowiedzilna za sprawdzenie czy plik zawiera klasyfikator
	 * i czy nie jest to multiklasyfikator, lub sprawdzenie podanych deklaracji
	 * w tym zakresie.
	 *
	 * @param file		java-owa scierzka do pliku.
	 * @param flags		flagi.
	 * @return			czy klasyfikator juz jest na listcie zarejestrowanych klasyfikatorow.
	 */
	private boolean testClassifier(String file, String name, int flags)
	{
		boolean toReturn = false;

		try {
			classClassifier = Class.forName(file);
			toReturn = DataManager.classifiersManager.containsClassifier(name);
			if ((flags & FLAG_TRY_TO_DETERMINE_FLAGS) == 0)
				if ((flags & FLAG_MULTICLASSIFIER) == 0)
				{
					Class superClass = classClassifier.getSuperclass();
					while ((superClass != Object.class) && (superClass != AbstractMultiClassifier.class) && (superClass != null))
						superClass = superClass.getSuperclass();
					if ((superClass == Object.class) || (superClass == null))
					{
						throw new RuntimeException("Klasa zadeklarowana jako multiklasifikator nie rozszerza klasy AbstractMultiClassifier");
					}
				}
				else
				{
					Class superClass = classClassifier.getSuperclass();
					while ((superClass != Object.class) && (superClass != ConfigurationWithStatistics.class) && (superClass != null))
						superClass = superClass.getSuperclass();
					if ((superClass == Object.class) || (superClass == null))
					{
						throw new RuntimeException("Klasa zadeklarowana jako klasifikator nie rozszerza klasy ConfigurationWithStatistics");
					}
				}
			else
			{
				Class superClass = classClassifier.getSuperclass();
				while ((superClass != Object.class) && (superClass != AbstractMultiClassifier.class) && (superClass != ConfigurationWithStatistics.class) && (superClass != null))
					superClass = superClass.getSuperclass();
				if (superClass == AbstractMultiClassifier.class)
				{
					flags |= FLAG_MULTICLASSIFIER;
					return toReturn;
				}
				if (superClass == ConfigurationWithStatistics.class)
				{
					flags &= ~FLAG_MULTICLASSIFIER;
					return toReturn;
				}
				throw new RuntimeException("Klasa nie rozszerza klasy AbstractClassifier ani AbstractMultiClassifier");
			}
		} catch (ClassNotFoundException e) {
			throw new RuntimeException("Cannot interprete classifier class as an object");
		}


		return toReturn;
	}

	/**
	 * wywolywane gdy nie ma zadnych plikow z domyslnymi opcjami
	 *
	 * @return	czy wszystko sie udalo
	 */
	public boolean noProperties()
	{
		if (state != CLASSIFIER_FILES_ADDED)
		{
			throw new RuntimeException("Uruchomiono dodawanie ustawien w nieodpowiednim momencie");
		}

		state = PROPERTY_FILES_ADDED;
		return true;
	}

	/**
	 * Najwazniejsze tesy klasyfikaratora.
	 * Jesli klasyfikator ma ustawione flagi to sprawdzamy,
	 * czy poradzi sobie z przykladowymi testami dla zadeklarowanych
	 * umiejetnosci klasyfiacji (patrz pierwsza lista stalych int-owych).
	 * Jesli klasyfikator nie ma ustawionych flag to probujemy sprawdzic
	 * dla jakich typow danych radzi sobie.
	 *
	 * Jesli wszystko pojdzie dobrze to klasyfikator zostanie dodany do
	 * listy wszystkich klasyfikatorow.
	 *
	 * @return	czy wszystko sie powiodlo
	 */
	public boolean addClassifier()
	{
		if (state != PROPERTY_FILES_ADDED)
		{
			throw new RuntimeException("Uruchomiono dodawanie klasyfikatora w nieodpowiednim momencie");
		}

		Report.displaynl("Testowanie mozliwosci klasyfikatora "+classClassifier.getName()+": ");

		if ((newFlags & FLAG_NUMERIC_TO_NUMERIC) != 0 || (newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) != 0)
		{
			try {
				throw new RuntimeException("nie zaimplementowane ");
				//practicalTest(classClassifier, newFlags, null);
				//newFlags |= FLAG_NUMERIC_TO_NUMERIC;
			} catch (Exception e) {
				if ((newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) == 0)
				{
					throw new RuntimeException("Blad. Nieudany test numeric to numeric dodawanego klasyfikatora");
				}
				else
					newFlags &= ~FLAG_NUMERIC_TO_NUMERIC;
			}
		}

		if ((newFlags & FLAG_NUMERIC_TO_NOMINAL) != 0 || (newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) != 0)
		{
			try {
				String[] arg = new String[2];
				arg[0] = "config" + File.separator + "vr" + File.separator + "system" + File.separator + "classifiersManagement" + File.separator + "daneTestowe" + File.separator + "iris.dat";
				arg[1] = "config" + File.separator + "vr" + File.separator + "system" + File.separator + "classifiersManagement" + File.separator + "daneTestowe" + File.separator + "iris.hdr";
				practicalTest(classClassifier, newFlags, arg);
				newFlags |= FLAG_NUMERIC_TO_NOMINAL;
				Report.displaynl("Klasyfikator obsluguje atrybuty numeryczne");
			} catch (Exception e) {
				if ((newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) == 0)
				{
					throw new RuntimeException("Blad. Nieudany test numeric to numeric dodawanego klasyfikatora");
				}
				else
				{
					newFlags &= ~FLAG_NUMERIC_TO_NOMINAL;
					Report.displaynl("Klasyfikator nie obsluguje atrybutow numerycznych");
				}
			}
		}

		if ((newFlags & FLAG_NOMINAL_TO_NUMERIC) != 0 || (newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) != 0)
		{
			try {
				throw new RuntimeException("nie zaimplementowane");
				//practicalTest(classClassifier, newFlags, null);
				//newFlags |= FLAG_NOMINAL_TO_NUMERIC;
			} catch (Exception e) {
				if ((newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) == 0)
				{
					throw new RuntimeException("Blad. Nieudany test numeric to numeric dodawanego klasyfikatora");
				}
				else
					newFlags &= ~FLAG_NOMINAL_TO_NUMERIC;
			}
		}

		if ((newFlags & FLAG_NOMINAL_TO_NOMINAL) != 0 || (newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) != 0)
		{
			try {
				String[] arg = new String[2];
				arg[0] = "config" + File.separator + "vr" + File.separator + "system" + File.separator + "classifiersManagement" + File.separator + "daneTestowe" + File.separator + "lymphography.dat";
				arg[1] = "config" + File.separator + "vr" + File.separator + "system" + File.separator + "classifiersManagement" + File.separator + "daneTestowe" + File.separator + "lymphography.hdr";
				practicalTest(classClassifier, newFlags, arg);
				newFlags |= FLAG_NOMINAL_TO_NOMINAL;
				Report.displaynl("Klasyfikator obsluguje atrybuty symboliczne");
			} catch (Exception e) {
				if ((newFlags & FLAG_TRY_TO_DETERMINE_FLAGS) == 0)
				{
					throw new RuntimeException("Blad. Nieudany test numeric to numeric dodawanego klasyfikatora");
				}
				else
				{
					newFlags &= ~FLAG_NOMINAL_TO_NOMINAL;
					Report.displaynl("Klasyfikator nie obsluguje atrybutow symbolicznych");
				}
			}
		}

		DataManager.classifiersManager.addTestedClassifier(classifierName, classifierJavaPathWithoutName, newFlags);

		state = ALL_DONE;
		return true;
	}

	/**
	 * Odpowiada za przetestowanie klasyfikatora na podanych danych
	 *
	 * @param classifier			obiekt typu Class reprezentujacy klasyfikator
	 * @param flags					flagi klasyfikatora
	 * @param args					argumenty (takie jak dla main z BasicTestTrainer)
	 */
	private void practicalTest(Class classifier, int flags, String[] args) throws Exception
	{
        // analiza argumentow i utworzenie naglowka
        DataFormatRecognizer rec = new DataFormatRecognizer();
        File dataFile = new File(args[0]);
        Header dataHeader = null;
        boolean oneFile = true;
        if (rec.isRses2xFormat(dataFile))
        {
            dataHeader = new ArrayHeader(dataFile);
            if (args.length==2) oneFile = false;
        }
        else
        {
            dataHeader = new ArrayHeader(new File(args[args.length-1]));
            if (args.length==3) oneFile = false;
        }

        // wczytanie tabel i podzial na treningowa i testowa czesc w przypadku pojedynczej tabeli
        DoubleDataTable trainTable = new ArrayListDoubleDataTable(dataFile, dataHeader, new EmptyProgress());
        DoubleDataTable testTable;
        if (oneFile)
        {
            Collection[] parts = trainTable.randomSplit(2, 1);
            trainTable = new ArrayListDoubleDataTable(parts[0]);
            testTable = new ArrayListDoubleDataTable(parts[1]);
        }
        else
        {
            testTable = new ArrayListDoubleDataTable(new File(args[1]), dataHeader, new EmptyProgress());
        }

		GeneralMultiClassifier tester = new GeneralMultiClassifier(null);
		tester.addNewClassifier(classifier, flags);
		tester.train(trainTable);
		tester.classify(testTable);
		if (tester.errors_appeared)
			throw new RuntimeException();

	}

}
