/*
 * Created on 2005-03-07
 *
 */
package vr.system.classifiersManagement;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import rseslib.structure.attribute.Attribute;
import rseslib.structure.attribute.Header;
import rseslib.system.Report;
import vr.system.DataManager;
import vr.system.options.SavableConfiguration;

/**
 * Klasa odpowiedzialna za zarzadzanie zarejestrowanymi klasyfikatorami:
 * 1. HeaderReader listy klasyfikatorow z pliku konfiguracyjnego
 * 2. Przekazywanie listy klasyfikatorow
 * 3. Dodawanie nowych klasyfikatorow do listy zarejestrowanych
 *
 * @author Marcin Staszczyk
 */
public class ClassifiersManager extends SavableConfiguration {

	/**
	 * Stale dostepowe do pliku properies opisujacego zarejestrowane klasyfikatory.
	 */
	private static final String NUMBER_OF_CLASSIFICATORS = "numberOfClassificators";
	private static final String FIRST_PART_OF_CLASSIFIERS_NAME = "classifier";
	private static final String LAST_PART_OF_CLASSIFIERS_NAME = "name";
	private static final String FIRST_PART_OF_CLASSIFIERS_DIR = "classifier";
	private static final String LAST_PART_OF_CLASSIFIERS_DIR = "dir";
	private static final String FIRST_PART_OF_CLASSIFIERS_FLAGS = "classifier";
	private static final String LAST_PART_OF_CLASSIFIERS_FLAGS = "flags";

	private Map nameToId;
	private Class[] classifiers;
	private ClassifierInfo[] classInfos;


	/**
	 * Konstruktor inicjujacy zmienne i czytujacy liste klasyfikatorow z pliku konfiguracyjnego.
	 */
	public ClassifiersManager() {
		super(null);
		nameToId = new HashMap();
		refreshClassifiersList();
	}

	/** Zwraca liczbe wszystkich zarejestrowanych klasyfikatorow */
	private int numberOfClassifiers()
	{
		return getIntProperty(NUMBER_OF_CLASSIFICATORS);
	}

	/**
	 * Odswieza liste klasyfikatorow na podstawie pliku konfiguracyjnego.
	 *
	 */
	public void refreshClassifiersList()
	{
		int classNum = getIntProperty(NUMBER_OF_CLASSIFICATORS);
		classifiers = new Class[classNum];
		classInfos = new ClassifierInfo[classNum];
		for (int i = 0; i < classNum; ++i)
		{
			try
			{
				String className = getProperty(FIRST_PART_OF_CLASSIFIERS_NAME + i + LAST_PART_OF_CLASSIFIERS_NAME);
				String classDir = getProperty(FIRST_PART_OF_CLASSIFIERS_DIR + i + LAST_PART_OF_CLASSIFIERS_DIR);
				int flags = getIntProperty(FIRST_PART_OF_CLASSIFIERS_FLAGS + i + LAST_PART_OF_CLASSIFIERS_FLAGS);
				try
				{
					classifiers[i] = Class.forName(classDir + className);
					classInfos[i] = new ClassifierInfo(className, flags);
					nameToId.put(className, new Integer(i));
				}
				catch (ClassNotFoundException e)
				{
					RuntimeException re = new RuntimeException("Cannot find class for classifier " + className + '\n' + e);
					Report.exception(re);
					throw re;
				}
			}
			catch (RuntimeException re)
				{}
		}
	}

	/**
	 * Metoda zwraca liste (tablice) klasyfikatorow ktorych flagi odpowiadaja aktualnie wczytanemu naglowkowi
	 * Kazdy zwracany klasyfikator (przynajmniej w teorii) nie powinien wywalic sie na danych zgodnych z aktualnym
	 * naglowkiem.
	 *
	 * @return Tablica informacji o klasyfikatorach: nazwa, flagi.
	 */
	public ClassifierInfo[] getClassifiersNames()
	{
		boolean numCond = false;
		boolean symbCond = false;
		boolean numDec = false;
		boolean symbDec = false;

		Header hdr = DataManager.dataHeader;
		for(int i = 0; i < hdr.noOfAttr(); ++i)
		{
			Attribute atr = hdr.attribute(i);
			if (atr.isConditional())
				if (atr.isNumeric())
					numCond = true;
				else
					symbCond = true;
			else
				if (atr.isNumeric())
					numDec = true;
				else
					symbDec = true;
		}

		int flags = 0;
		if ((numCond) && (numDec))
			flags += ClassifierRegistrator.FLAG_NUMERIC_TO_NUMERIC;
		if ((numCond) && (symbDec))
			flags += ClassifierRegistrator.FLAG_NUMERIC_TO_NOMINAL;
		if ((symbCond) && (numDec))
			flags += ClassifierRegistrator.FLAG_NOMINAL_TO_NUMERIC;
		if ((symbCond) && (symbDec))
			flags += ClassifierRegistrator.FLAG_NOMINAL_TO_NOMINAL;

		return getClassifiersNames(flags);
	}

	/**
	 * Dla podanych flag zwraca tablice klasyfikatorow spelniajacych podanych watunkow.
	 *
	 * @param flags	Zdolnosci wymagane od klasyfikatorow.
	 * @return		Tablica informacji o klasyfikatorach: nazwa, flagi.
	 */
	private ClassifierInfo[] getClassifiersNames(int flags)
	{
		ArrayList list = new ArrayList();

		for (int i = 0; i < classInfos.length; ++i)
			if (((classInfos[i] != null) && (classInfos[i].flags & flags) == flags))
				list.add(classInfos[i]);

		return (ClassifierInfo[])list.toArray(new ClassifierInfo[list.size()]);
	}

	/**
	 * Dla podanego identyfikatora zwraca odpowiadajacy mu obiekt typu Class.
	 *
	 * @param id	Identyfikator klasyfikatora.
	 * @return		Obiekt typu Class.
	 */
	private Class getClassifierClass(int id)
	{
		if ((id < classifiers.length) && (classifiers[id] != null))
			return classifiers[id];

		return null;
	}
	
	/**
	 * Dla podanej nazwy zwraca, czy istnije juz klasyfikator o tej nazwie.
	 * 
	 * @param name	Nazwa klasyfikatora.
	 * @return		czy klasyfikator juz jest na liscie.
	 */
	public boolean containsClassifier(String name)
	{
		return nameToId.containsKey(name);
	}

	/**
	 * Dla podanej nazwy zwraca odpowiadajacy jej obiekt typu Class.
	 *
	 * @param name	Nazwa klasyfikatora.
	 * @return		Obiekt typu Class.
	 */
	public Class getClassifierClass(String name)
	{
		if (nameToId.containsKey(name))
			return getClassifierClass(((Integer)nameToId.get(name)).intValue());
		return null;
	}

	/**
	 * Dodaje nowy klasyfikator do listy zarejestrowanych w pamieci i do pliku configuracyjnego
	 *
	 * @param name
	 * @param dir
	 * @param flags
	 * @throws RuntimeException
	 */
	public void addTestedClassifier(String name, String dir, int flags) throws RuntimeException
	{
		Class classifier;
		ClassifierInfo classInfo;
		int clNum;
		
		try {
			classifier = Class.forName(dir + name);
			classInfo = new ClassifierInfo(name, flags);
		} catch (ClassNotFoundException e) {
			throw new RuntimeException("BLAD: Dodawanie nieistniejacego klasyfikatora do listy klasyfikatorow");
		}
		
		if (nameToId.containsKey(name))
		{
			clNum = ((Integer)nameToId.get(name)).intValue();
			classifiers[clNum] = classifier;
			classInfos[clNum] = classInfo;
		}
		else
		{
			clNum = getIntProperty(NUMBER_OF_CLASSIFICATORS);
			setProperty(NUMBER_OF_CLASSIFICATORS, "" + (clNum+1));
			Class[] clCl = new Class[clNum+1];
			ClassifierInfo[] clInf = new ClassifierInfo[clNum+1];
			for (int i = 0; i < clNum; ++i)
			{
				clCl[i] = classifiers[i];
				clInf[i] = classInfos[i];
			}
			clCl[clNum] = classifier;
			clInf[clNum] = classInfo;
			nameToId.put(name, new Integer(clNum));	
			
			classifiers = clCl;
			classInfos = clInf;
		}
		
		setProperty(FIRST_PART_OF_CLASSIFIERS_NAME + clNum + LAST_PART_OF_CLASSIFIERS_NAME, name);
		setProperty(FIRST_PART_OF_CLASSIFIERS_DIR + clNum + LAST_PART_OF_CLASSIFIERS_DIR, dir);
		setProperty(FIRST_PART_OF_CLASSIFIERS_FLAGS + clNum + LAST_PART_OF_CLASSIFIERS_FLAGS, "" + flags);

		writeProperties();

		DataManager.chosenClassifiers = new ArrayList();
	}
}
