/*
 * Created on 2005-03-20
 *
 */
package vr.system.options;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Properties;

/**
 * The instances of this class
 * provides access to properties
 * provided in a constructor
 * or defined in the properties file.
 * The properties file name is the name of the class
 * extending this abstract class.
 * Class is quite exac copy of class Configuration
 *
 * @version     1.0 20 March 2005
 * @author      Grzegorz Gra, Arkadiusz Wojna and Marcin Staszczyk
 */
public abstract class SavableConfiguration
{
    /** SavableConfiguration directory path. */
    protected static final String CONFIGURATION_DIRECTORY = "config"+File.separator;
    /** Properties file extension. */
    protected static final String PROPERTIES_EXTENSION = ".properties";

    /** Map between property names and property values. */
    protected Properties m_Properties;
    /** path to file with properties */
    protected String path;

    /**
     * Constructor that reads properties from properties file.
     *
     * @param prop Map between property names and property values.
     */
    public SavableConfiguration(Properties prop)
    {
        if (prop!=null) m_Properties = (Properties)prop.clone();
        else m_Properties = loadDefaultProperties(getClass());
    }

    /**
     * Returns the default properties for a given class.
     *
     * @param configurableClass       Class.
     * @return Properties             Default properties for a given class.
     */
    public static Properties loadDefaultProperties(Class configurableClass)
    {
        Class superclass = configurableClass;
        while (superclass!=null && superclass!=SavableConfiguration.class) superclass = superclass.getSuperclass();
        if (superclass==null) throw new RuntimeException("Error while reading properties file: "+configurableClass.getName()+" not a subclass of rseslib.system.SavableConfiguration");
        Properties prop = new Properties();
        String[] pathes = getPropertiesFilePathes(configurableClass.getName());
        for (int path = 0; path < pathes.length; path++)
        {
            File propertyFile = new File(pathes[path]);
            try
            {
                if (propertyFile.exists()) prop.load(new BufferedInputStream(new FileInputStream(propertyFile)));
            }
            catch (IOException e)
            {
                throw new RuntimeException("Error while reading properties file: "+e.getMessage());
            }
        }
        return prop;
    }

    /**
     * Returns the pathes to the properties files for a given class name.
     *
     * @param className               Class name.
     * @return                        Pathes to the properties files for a given class name.
     * @throws ClassNotFoundException If the class is not found.
     */
    protected static String[] getPropertiesFilePathes(String className)
    {
        Class classReference = null;
        try
        {
            classReference = Class.forName(className);
        }
        catch (ClassNotFoundException e)
        {
            throw new RuntimeException("Class "+className+" not found");
        }
        ArrayList pathes = new ArrayList();
        while (classReference!=null)
        {
            className = classReference.getName();
            pathes.add(CONFIGURATION_DIRECTORY+className.replace('.', File.separatorChar)+PROPERTIES_EXTENSION);
            classReference = classReference.getSuperclass();
        }
        return (String[])pathes.toArray(new String[0]);
    }

    /**
     * Returns property value as a string.
     *
     * @param propertyName     Property name.
     * @return                 Property value.
     */
    public String getProperty(String propertyName)
    {
        String propertyValueString = m_Properties.getProperty(propertyName);
        if (propertyValueString==null) throw new RuntimeException("Property "+propertyName+" for the class "+this.getClass().getName()+" is not defined");
        return propertyValueString;
    }

    /**
     * Returns property value as a boolean.
     *
     * @param propertyName     Property name.
     * @return                 Property value.
     */
    public boolean getBoolProperty(String propertyName)
    {
        boolean val = false;
        String propertyValueString = m_Properties.getProperty(propertyName);
        if (propertyValueString==null) throw new RuntimeException("Property "+propertyName+" for the class "+this.getClass().getName()+" is not defined");
        if (propertyValueString.equalsIgnoreCase("true")) val = true;
        else if (propertyValueString.equalsIgnoreCase("false")) val = false;
        else throw new RuntimeException("The value of the property "+propertyName+" is neither TRUE nor FALSE");
        return val;
    }

    /**
     * Returns property value as int.
     *
     * @param propertyName     Property name.
     * @return                 Property value.
     */
    public int getIntProperty(String propertyName)
    {
        String propertyValueString = m_Properties.getProperty(propertyName);
        if (propertyValueString==null) throw new RuntimeException("Property "+propertyName+" for the class "+this.getClass().getName()+" is not defined");
        return Integer.parseInt(propertyValueString);
    }

    /**
     * Returns property value as double.
     *
     * @param propertyName     Property name.
     * @return                 Property value.
     */
    public double getDoubleProperty(String propertyName)
    {
        String propertyValueString = m_Properties.getProperty(propertyName);
        if (propertyValueString==null) throw new RuntimeException("Property "+propertyName+" for the class "+this.getClass().getName()+" is not defined");
        return Double.parseDouble(propertyValueString);
    }

    /**
     * Sets property to a given value.
     *
     * @param propertyName     Property name.
     * @param propertyValue    New property value.
     */
    public void setProperty(String propertyName, String propertyValue)
    {
        m_Properties.setProperty(propertyName, propertyValue);
    }

    /**
     * Writes properties through a given stream.
     *
     * @param arg	Stream.
     */
    public void writeProperties(PrintStream arg)
    {
        Enumeration itr = m_Properties.propertyNames();
        while (itr.hasMoreElements())
        {
        	String prop = (String)itr.nextElement();
        	String val = m_Properties.getProperty(prop);
        	arg.println(prop + " = " + val);
        }
        arg.flush();
    }

    /**
     * Write properties to default properties folder.
     *
     */
    protected void writeProperties()
    {
    	String className = this.getClass().getName();
        String propertiesFile = (CONFIGURATION_DIRECTORY+className.replace('.', File.separatorChar)+PROPERTIES_EXTENSION);
        File file = new File(propertiesFile);
        if (! file.getParentFile().exists())
        	file.getParentFile().mkdirs();
        if (! file.exists())
			try {
				file.createNewFile();
			} catch (IOException e1) {
				throw new RuntimeException("IOError");
			}

        try {
        	writeProperties(new PrintStream(new FileOutputStream(propertiesFile)));
		} catch (FileNotFoundException e) {
			throw new RuntimeException("Blad. Nie powinno byc problemow\n" + e);
		}
    }
}
